/*
 * Copyright (C) 2005-2010 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file audiod.c the paraslash's audio daemon */
#include <regex.h>
#include <sys/types.h>
#include <dirent.h>
#include <signal.h>
#include <openssl/rc4.h>
#include <stdbool.h>

#include "para.h"
#include "error.h"
#include "crypt.h"
#include "audiod.cmdline.h"
#include "list.h"
#include "sched.h"
#include "ggo.h"
#include "recv.h"
#include "buffer_tree.h"
#include "filter.h"
#include "grab_client.h"
#include "client.cmdline.h"
#include "client.h"
#include "audiod.h"
#include "net.h"
#include "daemon.h"
#include "string.h"
#include "fd.h"
#include "write.h"
#include "write_common.h"
#include "signal.h"

/** define the array of error lists needed by para_audiod */
INIT_AUDIOD_ERRLISTS;
/** define the array containing all supported audio formats */
const char *audio_formats[] = {AUDIOD_AUDIO_FORMAT_ARRAY NULL};

/** Defines how audiod handles one supported audio format. */
struct audio_format_info {
	/** pointer to the receiver for this audio format */
	struct receiver *receiver;
	/** the receiver configuration */
	void *receiver_conf;
	/** the number of filters that should be activated for this audio format */
	unsigned int num_filters;
	/** Array of filter numbers to be activated. */
	unsigned *filter_nums;
	/** Pointer to the array of filter configurations. */
	void **filter_conf;
	/** the number of filters that should be activated for this audio format */
	unsigned int num_writers;
	/** Array of writer numbers to be activated. */
	int *writer_nums;
	/** pointer to the array of writer configurations */
	void **writer_conf;
	/** do not start receiver/filters/writer before this time */
	struct timeval restart_barrier;
};

/**
 * para_audiod uses \p MAX_STREAM_SLOTS different slots, each of which may
 * be associated with a receiver/filter/writer triple. This array holds all
 * information on the status of these slots.
 *
 * \sa struct slot_info
 * */
struct slot_info slot[MAX_STREAM_SLOTS];

/** The vss status flags audiod is interested in. */
enum vss_status_flags {
	/** Whether the 'N' flag is set. */
	VSS_STATUS_FLAG_NEXT = 1,
	/** The 'P' flag is set. */
	VSS_STATUS_FLAG_PLAYING = 2,
};

/**
 * The task for obtaining para_server's status (para_client stat).
 *
 * \sa struct task, struct sched.
 */
struct status_task {
	/** The associated task structure of audiod. */
	struct task task;
	/** Client data associated with the stat task. */
	struct client_task *ct;
	/** Do not restart client command until this time. */
	struct timeval restart_barrier;
	/** Last time we received status data from para_server. */
	struct timeval last_status_read;
	size_t min_iqs;
	/** The offset value announced by para_server. */
	int offset_seconds;
	/** The length of the current audio file as announced by para_server. */
	int length_seconds;
	/** The start of the current stream from the view of para_server. */
	struct timeval server_stream_start;
	/** The average time deviation between para_server and para_audiod. */
	struct timeval sa_time_diff;
	/** Whether client time is ahead of server time. */
	int sa_time_diff_sign;
	/** The 'P' and the 'N' flags as announced by para_server. */
	enum vss_status_flags vss_status;
	/** Number of times the clock difference is to be checked. */
	unsigned clock_diff_count;
	/** When to start the next check for clock difference. */
	struct timeval clock_diff_barrier;
	/** Number of the audio format as announced by para_server. */
	int current_audio_format_num;
	/* The status task btrn is the child of the client task. */
	struct btr_node *btrn;
};

/** The array of status items sent by para_server. */
char *stat_item_values[NUM_STAT_ITEMS] = {NULL};

/**
 * the current mode of operation of which can be changed by the on/off/cycle
 * commands. It is either, AUDIOD_OFF, AUDIOD_ON or AUDIOD_STANDBY.
 */
int audiod_status = AUDIOD_ON;

/**
 * the gengetopt args_info struct that holds information on all command line
 * arguments
 */
struct audiod_args_info conf;

static char *socket_name;
static struct audio_format_info afi[NUM_AUDIO_FORMATS];

static struct signal_task signal_task_struct, *sig_task = &signal_task_struct;

static struct status_task status_task_struct;

/**
 * the task that calls the status command of para_server
 *
 * \sa struct status_task
 */
static struct status_task *stat_task = &status_task_struct;

/**
 * the task for handling audiod commands
 *
 * \sa struct task, struct sched
 */
struct command_task {
	/** the local listening socket */
	int fd;
	/** the associated task structure */
	struct task task;
};

/** iterate over all supported audio formats */
#define FOR_EACH_AUDIO_FORMAT(af) for (af = 0; af < NUM_AUDIO_FORMATS; af++)

/**
 * Get the audio format number.
 *
 * \param name The name of the audio format.
 *
 * \return The audio format number on success, -E_UNSUPPORTED_AUDIO_FORMAT if
 * \a name is not a supported audio format.
 */
int get_audio_format_num(const char *name)
{
	int i;

	while (para_isspace(*name))
		name++;
	FOR_EACH_AUDIO_FORMAT(i)
		if (!strcmp(name, audio_formats[i]))
			return i;
	return -E_UNSUPPORTED_AUDIO_FORMAT;
}

/**
 * Compute the play time based on information of the given slot.
 *
 * \param slot_num The slot number (negative means: no slot).
 *
 * This computes a string of the form "0:07 [3:33] (3%/3:40)" using information
 * from the status items received from para_server and the start time of the
 * (first) writer of the given slot.
 *
 * It has to to take into account that probably the stream was not started at
 * the beginning of the file, that the clock between the server and the client
 * host may differ and that playback of the stream was delayed, e.g. because
 * the prebuffer filter is used in the filter chain of the given slot.
 *
 * If no writer is active in the given slot, or \a slot_num is negative
 * (indicating that para_audiod runs in standby mode), an approximation based
 * only on the status items is computed and the returned string is prefixed
 * with "~".
 *
 * \return A string that must be freed by the caller.
 */
char *get_time_string(int slot_num)
{
	int ret, seconds = 0, length;
	struct timeval *tmp, sum, sss, /* server stream start */
		rstime, /* receiver start time */
		wstime, /* writer start time */
		wtime, /* now - writer start */
		rskip; /* receiver start - sss */
	struct slot_info *s = slot_num < 0? NULL : &slot[slot_num];
	char *msg;

	if (audiod_status == AUDIOD_OFF)
		goto empty;
	if (!(stat_task->vss_status & VSS_STATUS_FLAG_PLAYING)) {
		if (stat_task->length_seconds) /* paused */
			return NULL;
		goto empty; /* stopped */
	}
	if (audiod_status == AUDIOD_ON && !s)
		goto empty;
	/*
	 * Valid status items and playing, set length and tmp to the stream
	 * start. We use the slot info and fall back to the info from current
	 * status items if no slot info is available.
	 */
	length = stat_task->length_seconds;
	tmp = &stat_task->server_stream_start;
	if (s && s->wns) { /* writer active in this slot */
		btr_get_node_start(s->wns[0].btrn, &wstime);
		if (wstime.tv_sec != 0) { /* writer wrote something */
			if (s->server_stream_start.tv_sec == 0) {
				/* copy status info to slot */
				s->server_stream_start = stat_task->server_stream_start;
				s->offset_seconds = stat_task->offset_seconds;
				s->seconds_total = stat_task->length_seconds;
			}
			length = s->seconds_total;
			tmp = &s->server_stream_start;
		}
	}
	if (stat_task->sa_time_diff_sign > 0)
		tv_diff(tmp, &stat_task->sa_time_diff, &sss);
	else
		tv_add(tmp, &stat_task->sa_time_diff, &sss);
	if (!s || !s->wns) {
		struct timeval diff;
		tv_diff(now, &sss, &diff);
		seconds = diff.tv_sec + stat_task->offset_seconds;
		goto out;
	}
	tv_diff(now, &wstime, &wtime);
	//PARA_CRIT_LOG("offset %d\n", s->offset_seconds);
	seconds = s->offset_seconds;
	btr_get_node_start(s->receiver_node->btrn, &rstime);
	ret = tv_diff(&rstime, &sss, &rskip);
	if (ret > 0) { /* audiod was started in the middle of the stream */
		tv_add(&wtime, &rskip, &sum);
		seconds += sum.tv_sec;
	} else
		seconds += wtime.tv_sec;
out:
	seconds = PARA_MIN(seconds, length);
	seconds = PARA_MAX(seconds, 0);
	msg = make_message(
		"%s%d:%02d [%d:%02d] (%d%%/%d:%02d)",
		s? "" : "~",
		seconds / 60,
		seconds % 60,
		(length - seconds) / 60,
		(length - seconds) % 60,
		length? (seconds * 100 + length / 2) / length : 0,
		length / 60,
		length % 60
	);
	PARA_DEBUG_LOG("slot %d: %s\n", slot_num, msg);
	return msg;
empty:
	return para_strdup(NULL);
}

static int want_colors(void)
{
	if (conf.color_arg == color_arg_no)
		return 0;
	if (conf.color_arg == color_arg_yes)
		return 1;
	if (conf.logfile_given)
		return 0;
	return isatty(STDERR_FILENO);
}

static void parse_config_or_die(void)
{
	int ret;
	char *config_file;
	struct audiod_cmdline_parser_params params = {
		.override = 0,
		.initialize = 0,
		.check_required = 1,
		.check_ambiguity = 0,
		.print_errors = 1
	};

	if (conf.config_file_given)
		config_file = para_strdup(conf.config_file_arg);
	else {
		char *home = para_homedir();
		config_file = make_message("%s/.paraslash/audiod.conf", home);
		free(home);
	}
	ret = file_exists(config_file);
	if (conf.config_file_given && !ret) {
		PARA_EMERG_LOG("can not read config file %s\n", config_file);
		goto err;
	}
	if (ret)
		audiod_cmdline_parser_config_file(config_file, &conf, &params);
	free(config_file);
	daemon_set_loglevel(conf.loglevel_arg);
	return;
err:
	free(config_file);
	exit(EXIT_FAILURE);
}

static void setup_signal_handling(void)
{
	sig_task->fd = para_signal_init();
	PARA_INFO_LOG("signal pipe: fd %d\n", sig_task->fd);
	para_install_sighandler(SIGINT);
	para_install_sighandler(SIGTERM);
	para_install_sighandler(SIGHUP);
	para_sigaction(SIGPIPE, SIG_IGN);
}

static void clear_slot(int slot_num)
{
	struct slot_info *s = &slot[slot_num];

	PARA_INFO_LOG("clearing slot %d\n", slot_num);
	memset(s, 0, sizeof(struct slot_info));
	s->format = -1;
}

static void close_receiver(int slot_num)
{
	struct slot_info *s = &slot[slot_num];
	struct audio_format_info *a;

	if (s->format < 0 || !s->receiver_node)
		return;
	a = &afi[s->format];
	PARA_NOTICE_LOG("closing %s receiver in slot %d\n",
		audio_formats[s->format], slot_num);
	a->receiver->close(s->receiver_node);
	btr_free_node(s->receiver_node->btrn);
	free(s->receiver_node);
	s->receiver_node = NULL;
}

static void writer_cleanup(struct writer_node *wn)
{
	struct writer *w;

	if (!wn)
		return;
	w = writers + wn->writer_num;
	PARA_INFO_LOG("closing %s\n", writer_names[wn->writer_num]);
	w->close(wn);
	btr_free_node(wn->btrn);
}

static void close_writers(struct slot_info *s)
{
	struct audio_format_info *a;
	int i;

	if (s->format < 0)
		return;
	assert(s->wns);
	a = afi + s->format;
	if (a->num_writers == 0)
		writer_cleanup(s->wns);
	else {
		for (i = 0; i < a->num_writers; i++)
			writer_cleanup(s->wns + i);
	}
	free(s->wns);
	s->wns = NULL;
}

static void close_filters(struct slot_info *s)
{
	int i;
	struct audio_format_info *a = afi + s->format;
	if (a->num_filters == 0)
		return;
	for (i = 0; i < a->num_filters; i++) {
		struct filter_node *fn = s->fns + i;
		struct filter *f;

		if (!fn)
			continue;
		f = filters + fn->filter_num;
		if (f->close)
			f->close(fn);
		btr_free_node(fn->btrn);
	}
	free(s->fns);
	s->fns = NULL;
}

/*
 * Whenever a task commits suicide by returning from post_select with t->error
 * < 0, it also removes its btr node. We do exactly that to kill a running
 * task. Note that the scheduler checks t->error also _before_ each pre/post
 * select call, so the victim will never be scheduled again.
 */
static void kill_btrn(struct btr_node *btrn, struct task *t, int error)
{
	if (t->error < 0)
		return;
	t->error = error;
	btr_remove_node(btrn);
}

static void kill_all_decoders(int error)
{
	int i, j;

	FOR_EACH_SLOT(i) {
		struct slot_info *s = &slot[i];
		struct audio_format_info *a;
		if (s->format < 0)
			continue;
		a = afi + s->format;
		if (s->wns)
			for (j = 0; j < a->num_writers; j++)
				kill_btrn(s->wns[j].btrn, &s->wns[j].task, error);
		if (s->fns)
			for (j = 0; j < a->num_writers; j++)
				kill_btrn(s->fns[j].btrn, &s->wns[j].task, error);
		if (s->receiver_node)
			kill_btrn(s->receiver_node->btrn, &s->receiver_node->task,
				error);
	}
}

static int get_empty_slot(void)
{
	int i;
	struct slot_info *s;

	FOR_EACH_SLOT(i) {
		s = &slot[i];
		if (s->format < 0) {
			clear_slot(i);
			return i;
		}
		if (s->wns || s->receiver_node || s->fns)
			continue;
		clear_slot(i);
		return i;
	}
	return -E_NO_MORE_SLOTS;
}

/**
 * get the number of filters
 *
 * \param audio_format_num the number identifying the audio format
 *
 * \return the number of filters for the given audio format
 *
 * \sa struct filter;
 */
int num_filters(int audio_format_num)
{
	return afi[audio_format_num].num_filters;
}

static void open_filters(struct slot_info *s)
{
	struct audio_format_info *a = afi + s->format;
	struct filter_node *fn;
	int nf = a->num_filters;
	struct btr_node *parent;
	int i;

	if (nf == 0)
		return;
	PARA_INFO_LOG("opening %s filters\n", audio_formats[s->format]);
	assert(s->fns == NULL);
	s->fns = para_calloc(nf * sizeof(struct filter_node));
	parent = s->receiver_node->btrn;
	for (i = 0; i < nf; i++) {
		struct filter *f = filters + a->filter_nums[i];
		fn = s->fns + i;
		fn->filter_num = a->filter_nums[i];
		fn->conf = a->filter_conf[i];
		fn->task.pre_select = f->pre_select;
		fn->task.post_select = f->post_select;

		fn->btrn = btr_new_node(&(struct btr_node_description)
			EMBRACE(.name = f->name, .parent = parent,
				.handler = f->execute, .context = fn));

		f->open(fn);
		register_task(&fn->task);
		parent = fn->btrn;
		PARA_NOTICE_LOG("%s filter %d/%d (%s) started in slot %d\n",
			audio_formats[s->format], i,  nf, f->name, (int)(s - slot));
		sprintf(fn->task.status, "%s (slot %d)", f->name, (int)(s - slot));
	}
}

static void open_writers(struct slot_info *s)
{
	int i;
	struct audio_format_info *a = afi + s->format;
	struct writer_node *wn;
	struct btr_node *parent = s->fns[a->num_filters - 1].btrn;

	assert(s->wns == NULL);
	s->wns = para_calloc(PARA_MAX(1U, a->num_writers)
		* sizeof(struct writer_node));
	if (a->num_writers == 0)
		setup_writer_node(NULL, parent, s->wns);
	else {
		PARA_INFO_LOG("opening %s writers\n", audio_formats[s->format]);
		for (i = 0; i < a->num_writers; i++) {
			wn = s->wns + i;
			wn->conf = a->writer_conf[i];
			wn->writer_num = a->writer_nums[i];
			register_writer_node(wn, parent);
		}
	}
}

/* returns slot num on success */
static int open_receiver(int format)
{
	struct audio_format_info *a = &afi[format];
	struct slot_info *s;
	int ret, slot_num;
	struct receiver *r = a->receiver;
	struct receiver_node *rn;

	tv_add(now, &(struct timeval)EMBRACE(2, 0), &a->restart_barrier);
	ret = get_empty_slot();
	if (ret < 0)
		return ret;
	slot_num = ret;
	rn = para_calloc(sizeof(*rn));
	rn->receiver = r;
	rn->conf = a->receiver_conf;
	rn->btrn = btr_new_node(&(struct btr_node_description)
		EMBRACE(.name = r->name, .context = rn));
	ret = r->open(rn);
	if (ret < 0) {
		btr_free_node(rn->btrn);
		free(rn);
		return ret;
	}
	s = &slot[slot_num];
	s->format = format;
	s->receiver_node = rn;
	PARA_NOTICE_LOG("started %s: %s receiver in slot %d\n",
		audio_formats[format], r->name, slot_num);
	rn->task.pre_select = r->pre_select;
	rn->task.post_select = r->post_select;
	sprintf(rn->task.status, "%s receiver node", r->name);
	register_task(&rn->task);
	return slot_num;
}

static bool receiver_running(void)
{
	int i;
	long unsigned ss1 = stat_task->server_stream_start.tv_sec;

	FOR_EACH_SLOT(i) {
		struct slot_info *s = &slot[i];
		long unsigned ss2 = s->server_stream_start.tv_sec;

		if (!s->receiver_node)
			continue;
		if (s->receiver_node->task.error >= 0)
			return true;
		if (ss1 == ss2)
			return true;
	}
	return false;
}

/**
 * Return the root node of the current buffer tree.
 *
 * This is only used for stream grabbing.
 *
 * \return \p NULL if no slot is currently active. If more than one buffer tree
 * exists, the node corresponding to the most recently started receiver is
 * returned.
 */
struct btr_node *audiod_get_btr_root(void)
{
	int i, newest_slot = -1;
	struct timeval newest_rstime = {0, 0};

	FOR_EACH_SLOT(i) {
		struct slot_info *s = &slot[i];
		struct timeval rstime;
		if (!s->receiver_node)
			continue;
		if (s->receiver_node->task.error < 0)
			continue;
		btr_get_node_start(s->receiver_node->btrn, &rstime);
		if (newest_slot >= 0 && tv_diff(&rstime, &newest_rstime, NULL) < 0)
			continue;
		newest_rstime = rstime;
		newest_slot = i;
	}
	if (newest_slot == -1)
		return NULL;
	return slot[newest_slot].receiver_node->btrn;
}

/* whether a new instance of a decoder should be started. */
static bool must_start_decoder(void)
{
	int cafn = stat_task->current_audio_format_num;
	unsigned vs = stat_task->vss_status;

	if (audiod_status != AUDIOD_ON)
		return false;
	if (cafn < 0)
		return false;
	if (!stat_task->ct)
		return false;
	if (vs & VSS_STATUS_FLAG_NEXT)
		return false;
	if (!(vs & VSS_STATUS_FLAG_PLAYING))
		return false;
	if (receiver_running())
		return false;
	if (tv_diff(now, &afi[cafn].restart_barrier, NULL) < 0)
		return false;
	return true;
}

static unsigned compute_time_diff(const struct timeval *status_time)
{
	struct timeval tmp, diff;
	static unsigned count;
	int sign, sa_time_diff_sign = stat_task->sa_time_diff_sign;
	const struct timeval max_deviation = {0, 500 * 1000};
	const int time_smooth = 5;

	if (!status_time)
		return count;
	sign = tv_diff(status_time, now, &diff);
//		PARA_NOTICE_LOG("%s: sign = %i, sa_time_diff_sign = %i\n", __func__,
//			sign, sa_time_diff_sign);
	if (!count) {
		sa_time_diff_sign = sign;
		stat_task->sa_time_diff = diff;
		count++;
		goto out;
	}
	if (count > 5) {
		int s = tv_diff(&diff, &stat_task->sa_time_diff, &tmp);
		if (tv_diff(&max_deviation, &tmp, NULL) < 0)
			PARA_WARNING_LOG("time diff jump: %lims\n",
				s * tv2ms(&tmp));
	}
	count++;
	sa_time_diff_sign = tv_convex_combination(
		sa_time_diff_sign * time_smooth, &stat_task->sa_time_diff,
		count > 10? sign : sign * time_smooth, &diff,
		&tmp);
	stat_task->sa_time_diff = tmp;
	PARA_INFO_LOG("time diff (cur/avg): %s%lums/%s%lums\n",
		sign < 0? "-" : "+",
		tv2ms(&diff),
		sa_time_diff_sign < 0? "-" : "+",
		tv2ms(&stat_task->sa_time_diff)
	);
out:
	stat_task->sa_time_diff_sign = sa_time_diff_sign;
	return count;
}

static int update_item(int itemnum, char *buf)
{
	long unsigned sec, usec;

	if (stat_task->clock_diff_count && itemnum != SI_CURRENT_TIME)
		return 1;
	free(stat_item_values[itemnum]);
	stat_item_values[itemnum] = para_strdup(buf);
	stat_client_write_item(itemnum);
	switch (itemnum) {
	case SI_STATUS_FLAGS:
		stat_task->vss_status = 0;
		if (strchr(buf, 'N'))
			stat_task->vss_status |= VSS_STATUS_FLAG_NEXT;
		if (strchr(buf, 'P'))
			stat_task->vss_status |= VSS_STATUS_FLAG_PLAYING;
		break;
	case SI_OFFSET:
		stat_task->offset_seconds = atoi(buf);
		break;
	case SI_SECONDS_TOTAL:
		stat_task->length_seconds = atoi(buf);
		break;
	case SI_STREAM_START:
		if (sscanf(buf, "%lu.%lu", &sec, &usec) == 2) {
			stat_task->server_stream_start.tv_sec = sec;
			stat_task->server_stream_start.tv_usec = usec;
		}
		break;
	case SI_CURRENT_TIME:
		if (sscanf(buf, "%lu.%lu", &sec, &usec) == 2) {
			struct timeval tv = {sec, usec};
			compute_time_diff(&tv);
		}
		break;
	case SI_FORMAT:
		stat_task->current_audio_format_num
			= get_audio_format_num(buf);
	}
	return 1;
}

static int parse_stream_command(const char *txt, char **cmd)
{
	char *p = strchr(txt, ':');
	int i;

	if (!p)
		return -E_MISSING_COLON;
	p++;
	FOR_EACH_AUDIO_FORMAT(i) {
		if (strncmp(txt, audio_formats[i], strlen(audio_formats[i])))
			continue;
		*cmd = p;
		return i;
	}
	return -E_UNSUPPORTED_AUDIO_FORMAT;
}

static int add_filter(int format, char *cmdline)
{
	struct audio_format_info *a = &afi[format];
	int filter_num, nf = a->num_filters;

	filter_num = check_filter_arg(cmdline, &a->filter_conf[nf]);
	if (filter_num < 0)
		return filter_num;
	a->filter_nums[nf] = filter_num;
	a->num_filters++;
	PARA_INFO_LOG("%s filter %d: %s\n", audio_formats[format], nf,
		filters[filter_num].name);
	return filter_num;
}

static int parse_writer_args(void)
{
	int i, ret, nw;
	char *cmd;
	struct audio_format_info *a;

	nw = PARA_MAX(1U, conf.writer_given);
	PARA_INFO_LOG("maximal number of writers: %d\n", nw);
	FOR_EACH_AUDIO_FORMAT(i) {
		a = &afi[i];
		a->writer_conf = para_malloc(nw * sizeof(void *));
		a->writer_nums = para_malloc(nw * sizeof(int));
		a->num_writers = 0;
	}
	for (i = 0; i < conf.writer_given; i++) {
		void *wconf;
		int writer_num;
		ret = parse_stream_command(conf.writer_arg[i], &cmd);
		if (ret < 0)
			goto out;
		a = &afi[ret];
		nw = a->num_writers;
		wconf = check_writer_arg(cmd, &writer_num);
		if (!wconf) {
			ret = writer_num;
			goto out;
		}
		a->writer_nums[nw] = writer_num;
		a->writer_conf[nw] = wconf;
		PARA_INFO_LOG("%s writer #%d: %s\n", audio_formats[ret],
			nw, writer_names[writer_num]);
		a->num_writers++;
	}
	ret = 1;
out:
	return ret;
}

static int parse_receiver_args(void)
{
	int i, ret, receiver_num;
	char *cmd = NULL;
	struct audio_format_info *a;

	for (i = conf.receiver_given - 1; i >= 0; i--) {
		char *arg = conf.receiver_arg[i];
		char *recv_arg = strchr(arg, ':');
		ret = -E_MISSING_COLON;
		if (!recv_arg)
			goto out;
		*recv_arg = '\0';
		recv_arg++;
		ret = get_audio_format_num(arg);
		if (ret < 0)
			goto out;
		afi[ret].receiver_conf = check_receiver_arg(recv_arg, &receiver_num);
		if (!afi[ret].receiver_conf) {
			ret = -E_RECV_SYNTAX;
			goto out;
		}
		afi[ret].receiver = &receivers[receiver_num];
	}
	/* use the first available receiver with no arguments
	 * for those audio formats for which no receiver
	 * was specified
	 */
	cmd = para_strdup(receivers[0].name);
	FOR_EACH_AUDIO_FORMAT(i) {
		a = &afi[i];
		if (a->receiver_conf)
			continue;
		a->receiver_conf = check_receiver_arg(cmd, &receiver_num);
		if (!a->receiver_conf)
			return -E_RECV_SYNTAX;
		a->receiver = &receivers[receiver_num];
	}
	ret = 1;
out:
	free(cmd);
	return ret;
}

static int init_default_filters(void)
{
	int i, ret = 1;

	FOR_EACH_AUDIO_FORMAT(i) {
		struct audio_format_info *a = &afi[i];
		char *tmp;
		int j;

		if (a->num_filters)
			continue; /* no default -- nothing to to */
		/*
		 * If udp is used to receive this audiod format, add fecdec as
		 * the first filter.
		 */
		if (strcmp(afi[i].receiver->name, "udp") == 0) {
			tmp = para_strdup("fecdec");
			add_filter(i, tmp);
			free(tmp);
			if (ret < 0)
				goto out;
		}
		/* add "dec" to audio format name */
		tmp = make_message("%sdec", audio_formats[i]);
		for (j = 0; filters[j].name; j++)
			if (!strcmp(tmp, filters[j].name))
				break;
		free(tmp);
		ret = -E_UNSUPPORTED_FILTER;
		if (!filters[j].name)
			goto out;
		tmp = para_strdup(filters[j].name);
		ret = add_filter(i, tmp);
		free(tmp);
		if (ret < 0)
			goto out;
		PARA_INFO_LOG("%s -> default filter: %s\n", audio_formats[i],
			filters[j].name);
	}
out:
	return ret;
}

static int parse_filter_args(void)
{
	int i, ret, nf;

	nf = PARA_MAX(2U, conf.filter_given);
	PARA_INFO_LOG("maximal number of filters: %d\n", nf);
	FOR_EACH_AUDIO_FORMAT(i) {
		afi[i].filter_conf = para_malloc(nf * sizeof(void *));
		afi[i].filter_nums = para_malloc(nf * sizeof(unsigned));
	}
	if (!conf.no_default_filters_given)
		return init_default_filters();
	for (i = 0; i < conf.filter_given; i++) {
		char *arg = conf.filter_arg[i];
		char *filter_name = strchr(arg, ':');
		ret = -E_MISSING_COLON;
		if (!filter_name)
			goto out;
		*filter_name = '\0';
		filter_name++;
		ret = get_audio_format_num(arg);
		if (ret < 0)
			goto out;
		ret = add_filter(ret, filter_name);
		if (ret < 0)
			goto out;
	}
	ret = init_default_filters(); /* use default values for the rest */
out:
	return ret;
}

static int parse_stream_args(void)
{
	int ret;

	ret = parse_receiver_args();
	if (ret < 0)
		return ret;
	ret = parse_filter_args();
	if (ret < 0)
		return ret;
	ret = parse_writer_args();
	if (ret < 0)
		return ret;
	return 1;
}

/* does not unlink socket on errors */
static int audiod_get_socket(void)
{
	struct sockaddr_un unix_addr;
	int ret, fd;

	if (conf.socket_given)
		socket_name = para_strdup(conf.socket_arg);
	else {
		char *hn = para_hostname();
		socket_name = make_message("/var/paraslash/audiod_socket.%s",
			hn);
		free(hn);
	}
	PARA_NOTICE_LOG("local socket: %s\n", socket_name);
	if (conf.force_given)
		unlink(socket_name);
	ret = create_local_socket(socket_name, &unix_addr,
		S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IWOTH);
	if (ret < 0)
		goto err;
	fd = ret;
	if (listen(fd , 5) < 0) {
		ret = -ERRNO_TO_PARA_ERROR(errno);
		goto err;
	}
	ret = mark_fd_nonblocking(fd);
	if (ret < 0)
		goto err;
	return fd;
err:
	PARA_EMERG_LOG("%s\n", para_strerror(-ret));
	exit(EXIT_FAILURE);
}

static void signal_pre_select(struct sched *s, struct task *t)
{
	struct signal_task *st = container_of(t, struct signal_task, task);
	para_fd_set(st->fd, &s->rfds, &s->max_fileno);
}

static void signal_post_select(struct sched *s, struct task *t)
{
	struct signal_task *st = container_of(t, struct signal_task, task);

	if (!FD_ISSET(st->fd, &s->rfds))
		return;

	st->signum = para_next_signal();
	switch (st->signum) {
	case SIGINT:
	case SIGTERM:
	case SIGHUP:
		PARA_EMERG_LOG("terminating on signal %d\n", st->signum);
		clean_exit(EXIT_FAILURE, "caught deadly signal");
	}
}

static void signal_setup_default(struct signal_task *st)
{
	st->task.pre_select = signal_pre_select;
	st->task.post_select = signal_post_select;
	sprintf(st->task.status, "signal task");
}

static void command_pre_select(struct sched *s, struct task *t)
{
	struct command_task *ct = container_of(t, struct command_task, task);
	para_fd_set(ct->fd, &s->rfds, &s->max_fileno);
}

static void command_post_select(struct sched *s, struct task *t)
{
	int ret;
	struct command_task *ct = container_of(t, struct command_task, task);
	static struct timeval last_status_dump;
	struct timeval tmp, delay = {0, 500 * 1000};

	tv_add(&last_status_dump, &delay, &tmp);
	if (tv_diff(&tmp, now, NULL) < 0) {
		audiod_status_dump();
		last_status_dump = *now;
	}

	if (!FD_ISSET(ct->fd, &s->rfds))
		return;
	ret = handle_connect(ct->fd);
	if (ret < 0)
		PARA_ERROR_LOG("%s\n", para_strerror(-ret));
	audiod_status_dump();
}

static void init_command_task(struct command_task *ct)
{
	ct->task.pre_select = command_pre_select;
	ct->task.post_select = command_post_select;
	ct->task.error = 0;
	ct->fd = audiod_get_socket(); /* doesn't return on errors */
	sprintf(ct->task.status, "command task");
}

static void close_stat_pipe(void)
{
	if (!stat_task->ct)
		return;
	btr_free_node(stat_task->ct->btrn);
	client_close(stat_task->ct);
	stat_task->ct = NULL;
	clear_and_dump_items();
	stat_task->length_seconds = 0;
	stat_task->offset_seconds = 0;
	stat_task->vss_status = 0;
	stat_task->current_audio_format_num = -1;
	audiod_status_dump();
}

/**
 * close the connection to para_server and exit
 *
 * \param status the exit status which is passed to exit(3)
 * \param msg the log message
 *
 * Log \a msg with loglevel \p EMERG, close the connection to para_server if
 * open, and call \p exit(status). \a status should be either EXIT_SUCCESS or
 * EXIT_FAILURE.
 *
 * \sa exit(3)
 */
void __noreturn clean_exit(int status, const char *msg)
{
	PARA_EMERG_LOG("%s\n", msg);
	if (socket_name)
		unlink(socket_name);
	close_stat_pipe();
	exit(status);
}

/* avoid busy loop if server is down */
static void set_stat_task_restart_barrier(unsigned seconds)
{
	struct timeval delay = {seconds, 0};
	tv_add(now, &delay, &stat_task->restart_barrier);
}

static void try_to_close_slot(int slot_num)
{
	struct slot_info *s = &slot[slot_num];
	struct audio_format_info *a = afi + s->format;
	int i;

	if (s->format < 0)
		return;
	if (s->receiver_node && s->receiver_node->task.error != -E_TASK_UNREGISTERED)
		return;
	for (i = 0; i < a->num_filters; i++)
		if (s->fns && s->fns[i].task.error != -E_TASK_UNREGISTERED)
			return;
	if (a->num_writers > 0) {
		for (i = 0; i < a->num_writers; i++)
			if (s->wns && s->wns[i].task.error != -E_TASK_UNREGISTERED)
				return;
	} else {
		if (s->wns && s->wns[0].task.error != -E_TASK_UNREGISTERED)
			return;
	}
	PARA_INFO_LOG("closing slot %d\n", slot_num);
	close_writers(s);
	close_filters(s);
	close_receiver(slot_num);
	clear_slot(slot_num);
}

/*
 * Check if any receivers/filters/writers need to be started and do so if
 * necessary.
 */
static void start_stop_decoders(void)
{
	int i, ret;
	struct slot_info *sl;
	struct audio_format_info *a;

	FOR_EACH_SLOT(i)
		try_to_close_slot(i);
	if (audiod_status != AUDIOD_ON ||
			!(stat_task->vss_status & VSS_STATUS_FLAG_PLAYING))
		return kill_all_decoders(-E_NOT_PLAYING);
	if (!must_start_decoder())
		return;
	ret = open_receiver(stat_task->current_audio_format_num);
	if (ret < 0) {
		PARA_ERROR_LOG("%s\n", para_strerror(-ret));
		return;
	}
	sl = slot + ret;
	a = afi + sl->format;
	if (a->num_filters)
		open_filters(sl);
	open_writers(sl);
	activate_grab_clients();
	btr_log_tree(sl->receiver_node->btrn, LL_NOTICE);
}

static void status_pre_select(struct sched *s, struct task *t)
{
	struct status_task *st = container_of(t, struct status_task, task);
	int ret, cafn = stat_task->current_audio_format_num;

	if (must_start_decoder())
		goto min_delay;
	ret = btr_node_status(st->btrn, 0, BTR_NT_LEAF);
	if (ret > 0)
		goto min_delay;
	if (st->ct && audiod_status == AUDIOD_OFF)
		goto min_delay;
	if (!st->ct && audiod_status != AUDIOD_OFF)
		sched_request_barrier_or_min_delay(&st->restart_barrier, s);
	if (cafn >= 0)
		sched_request_barrier(&afi[cafn].restart_barrier, s);
	/*
	 * If para_server is playing we'd like to have a smooth time display
	 * even if we are running in standby mode. So we request a timeout that
	 * expires at the next full second.
	 */
	if (stat_task->vss_status & VSS_STATUS_FLAG_PLAYING)
		sched_request_timeout_ms(1000 - now->tv_usec / 1000, s);
	return;
min_delay:
	sched_min_delay(s);
}

/* restart the client task if necessary */
static void status_post_select(__a_unused struct sched *s, struct task *t)
{
	struct status_task *st = container_of(t, struct status_task, task);

	if (audiod_status == AUDIOD_OFF) {
		if (!st->ct)
			goto out;
		if (st->ct->task.error >= 0) {
			kill_btrn(st->ct->btrn, &st->ct->task, -E_AUDIOD_OFF);
			goto out;
		}
		if (st->ct->task.error != -E_TASK_UNREGISTERED)
			goto out;
		close_stat_pipe();
		st->clock_diff_count = conf.clock_diff_count_arg;
		goto out;
	}
	if (st->ct) {
		char *buf;
		size_t sz;
		int ret;
		if (st->ct->task.error < 0) {
			if (st->ct->task.error != -E_TASK_UNREGISTERED)
				goto out;
			close_stat_pipe();
			goto out;
		}
		if (st->ct->status != CL_RECEIVING)
			goto out;
		ret = btr_node_status(st->btrn, st->min_iqs, BTR_NT_LEAF);
		if (ret <= 0) {
			struct timeval diff;
			tv_diff(now, &st->last_status_read, &diff);
			if (diff.tv_sec > 61)
				kill_btrn(st->ct->btrn, &st->ct->task,
					-E_STATUS_TIMEOUT);
			goto out;
		}
		btr_merge(st->btrn, st->min_iqs);
		sz = btr_next_buffer(st->btrn, &buf);
		ret = for_each_stat_item(buf, sz, update_item);
		if (ret < 0) {
			kill_btrn(st->ct->btrn, &st->ct->task, ret);
			goto out;
		}
		if (sz != ret) {
			btr_consume(st->btrn, sz - ret);
			st->last_status_read = *now;
			st->min_iqs = 0;
		} else /* current status item crosses buffers */
			st->min_iqs = sz + 1;
		goto out;
	}
	if (tv_diff(now, &st->restart_barrier, NULL) < 0)
		goto out;
	if (st->clock_diff_count) { /* get status only one time */
		char *argv[] = {"audiod", "--", "stat", "-p", "-n=1", NULL};
		int argc = 5;
		PARA_INFO_LOG("clock diff count: %d\n", st->clock_diff_count);
		st->clock_diff_count--;
		client_open(argc, argv, &st->ct, NULL, NULL, st->btrn);
		set_stat_task_restart_barrier(2);

	} else {
		char *argv[] = {"audiod", "--", "stat", "-p", NULL};
		int argc = 4;
		client_open(argc, argv, &st->ct, NULL, NULL, st->btrn);
		set_stat_task_restart_barrier(5);
	}
	free(stat_item_values[SI_BASENAME]);
	stat_item_values[SI_BASENAME] = para_strdup(
		"no connection to para_server");
	stat_client_write_item(SI_BASENAME);
	st->last_status_read = *now;
out:
	start_stop_decoders();
}

static void init_status_task(struct status_task *st)
{
	memset(st, 0, sizeof(struct status_task));
	st->task.pre_select = status_pre_select;
	st->task.post_select = status_post_select;
	st->sa_time_diff_sign = 1;
	st->clock_diff_count = conf.clock_diff_count_arg;
	st->current_audio_format_num = -1;
	sprintf(st->task.status, "stat");
	st->btrn = btr_new_node(&(struct btr_node_description)
		EMBRACE(.name = "stat"));
}

static void set_initial_status(void)
{
	audiod_status = AUDIOD_ON;
	if (!conf.mode_given)
		return;
	if (!strcmp(conf.mode_arg, "sb")) {
		audiod_status = AUDIOD_STANDBY;
		return;
	}
	if (!strcmp(conf.mode_arg, "off")) {
		audiod_status = AUDIOD_OFF;
		return;
	}
	if (strcmp(conf.mode_arg, "on"))
		PARA_WARNING_LOG("invalid mode\n");
}

__noreturn static void print_help_and_die(void)
{
	int d = conf.detailed_help_given;
	const char **p = d? audiod_args_info_detailed_help
		: audiod_args_info_help;

	printf_or_die("%s\n\n", AUDIOD_CMDLINE_PARSER_PACKAGE "-"
		AUDIOD_CMDLINE_PARSER_VERSION);
	printf_or_die("%s\n\n", audiod_args_info_usage);
	for (; *p; p++)
		printf_or_die("%s\n", *p);
	print_receiver_helps(d);
	print_filter_helps(d);
	print_writer_helps(d);
	exit(0);
}

static void init_colors_or_die(void)
{
	int ret, i;

	if (!want_colors())
		return;
	daemon_set_default_log_colors();
	daemon_set_flag(DF_COLOR_LOG);
	for (i = 0; i < conf.log_color_given; i++) {
		ret = daemon_set_log_color(conf.log_color_arg[i]);
		if (ret < 0)
			exit(EXIT_FAILURE);
	}
}

/**
 * the main function of para_audiod
 *
 * \param argc usual argument count
 * \param argv usual argument vector
 *
 * \return EXIT_SUCCESS or EXIT_FAILURE
 *
 * \sa para_audiod(1)
 * */
int main(int argc, char *argv[])
{
	int ret, i;
	static struct sched s;
	struct command_task command_task_struct, *cmd_task = &command_task_struct;
	struct audiod_cmdline_parser_params params = {
		.override = 0,
		.initialize = 1,
		.check_required = 0,
		.check_ambiguity = 0,
		.print_errors = 1
	};

	valid_fd_012();
	if (audiod_cmdline_parser_ext(argc, argv, &conf, &params))
		exit(EXIT_FAILURE);
	HANDLE_VERSION_FLAG("audiod", conf);
	/* init receivers/filters/writers early to make help work */
	recv_init();
	filter_init();
	writer_init();
	if (conf.help_given || conf.detailed_help_given)
		print_help_and_die();
	drop_privileges_or_die(conf.user_arg, conf.group_arg);
	parse_config_or_die();
	init_colors_or_die();
	init_random_seed_or_die();
	daemon_set_flag(DF_LOG_TIME);
	daemon_set_flag(DF_LOG_HOSTNAME);
	daemon_set_flag(DF_LOG_LL);
	if (conf.log_timing_given)
		daemon_set_flag(DF_LOG_TIMING);
	if (conf.logfile_given) {
		daemon_set_logfile(conf.logfile_arg);
		daemon_open_log_or_die();
	}
	ret = parse_stream_args();
	if (ret < 0) {
		PARA_EMERG_LOG("%s\n", para_strerror(-ret));
		exit(EXIT_FAILURE);
	}
	log_welcome("para_audiod");
	server_uptime(UPTIME_SET);
	set_initial_status();
	FOR_EACH_SLOT(i)
		clear_slot(i);
	setup_signal_handling();
	signal_setup_default(sig_task);

	init_status_task(stat_task);
	init_command_task(cmd_task);

	if (conf.daemon_given)
		daemonize();

	register_task(&sig_task->task);
	register_task(&cmd_task->task);
	register_task(&stat_task->task);
	s.default_timeout.tv_sec = 2;
	s.default_timeout.tv_usec = 999 * 1000;
	ret = schedule(&s);

	PARA_EMERG_LOG("%s\n", para_strerror(-ret));
	return EXIT_FAILURE;
}
