/*
 * Copyright (C) 2005-2010 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file net.c Networking-related helper functions. */

/*
 * Since glibc 2.8, the _GNU_SOURCE feature test macro must be defined in order
 * to obtain the definition of the ucred structure.
 */
#define _GNU_SOURCE

#include <netdb.h>

/* At least NetBSD needs these. */
#ifndef AI_V4MAPPED
#define AI_V4MAPPED 0
#endif
#ifndef AI_ALL
#define AI_ALL 0
#endif
#ifndef AI_ADDRCONFIG
#define AI_ADDRCONFIG 0
#endif

#include <dirent.h>
#include <regex.h>
#include <openssl/rc4.h>

#include "para.h"
#include "error.h"
#include "crypt.h"
#include "net.h"
#include "string.h"
#include "list.h"
#include "fd.h"

/**
 * Parse and validate IPv4 address/netmask string.
 *
 * \param cidr	  Address in CIDR notation
 * \param addr	  Copy of the IPv4 address part of \a cidr
 * \param addrlen Size of \a addr in bytes
 * \param netmask Value of the netmask part in \a cidr or the
 *		  default of 32 if not specified.
 *
 * \return Pointer to \a addr if succesful, NULL on error.
 * \sa RFC 4632
 */
char *parse_cidr(const char *cidr,
		 char	 *addr, ssize_t addrlen,
		 int32_t *netmask)
{
	const char *o = cidr;
	char *c = addr, *end = c + (addrlen - 1);

	*netmask = 0x20;

	if (cidr == NULL || addrlen < 1)
		goto failed;

	for (o = cidr; (*c = *o == '/'? '\0' : *o); c++, o++)
		if (c == end)
			goto failed;

	if (*o == '/')
		if (para_atoi32(++o, netmask) < 0 ||
		    *netmask < 0 || *netmask > 0x20)
			goto failed;

	if (is_valid_ipv4_address(addr))
		return addr;
failed:
	*addr = '\0';
	return NULL;
}


/**
 * Match string as a candidate IPv4 address.
 *
 * \param address The string to match.
 * \return True if \a address has "dot-quad" format.
 */
static bool is_v4_dot_quad(const char *address)
{
	bool result;
	regex_t r;

	assert(para_regcomp(&r, "^([0-9]+\\.){3}[0-9]+$",
		REG_EXTENDED | REG_NOSUB) >= 0);
	result = regexec(&r, address, 0, NULL, 0) == 0;
	regfree(&r);
	return result;
}

/**
 * Perform basic syntax checking on the host-part of an URL:
 *
 * - Since ':' is invalid in IPv4 addresses and DNS names, the
 *   presence of ':' causes interpretation as IPv6 address;
 * - next the first-match-wins algorithm from RFC 3986 is applied;
 * - else the string is considered as DNS name, to be resolved later.
 *
 * \param host The host string to check.
 * \return True if \a host passes the syntax checks.
 *
 * \sa RFC 3986, 3.2.2; RFC 1123, 2.1; RFC 1034, 3.5
 */
static bool host_string_ok(const char *host)
{
	if (host == NULL || *host == '\0')
		return false;
	if (strchr(host, ':') != NULL)
		return is_valid_ipv6_address(host);
	if (is_v4_dot_quad(host))
		return is_valid_ipv4_address(host);
	return true;
}

/**
 * Parse and validate URL string.
 *
 * The URL syntax is loosely based on RFC 3986, supporting one of
 * - "["host"]"[:port] for native IPv6 addresses and
 * - host[:port] for IPv4 hostnames and DNS names.
 *
 * Native IPv6 addresses must be enclosed in square brackets, since
 * otherwise there is an ambiguity with the port separator `:'.
 * The 'port' part is always considered to be a number; if absent,
 * it is set to -1, to indicate that a default port is to be used.
 *
 * The following are valid examples:
 * - 10.10.1.1
 * - 10.10.1.2:8000
 * - localhost
 * - localhost:8001
 * - [::1]:8000
 * - [badc0de::1]
 *
 * \param url	  The URL string to take apart.
 * \param host	  To return the copied host part of \a url.
 * \param hostlen The maximum length of \a host.
 * \param port	  To return the port number (if any) of \a url.
 *
 * \return Pointer to \a host, or NULL if failed.
 * If NULL is returned, \a host and \a portnum are undefined. If no
 * port number was present in \a url, \a portnum is set to -1.
 *
 * \sa RFC 3986, 3.2.2/3.2.3
 */
char *parse_url(const char *url,
		char	*host, ssize_t hostlen,
		int32_t	*port)
{
	const char *o = url;
	char *c = host, *end = c + (hostlen - 1);

	*port = -1;

	if (o == NULL || hostlen < 1)
		goto failed;

	if (*o == '[') {
		for (++o; (*c = *o == ']' ? '\0' : *o); c++, o++)
			if (c == end)
				goto failed;

		if (*o++ != ']' || (*o != '\0' && *o != ':'))
			goto failed;
	} else {
		for (; (*c = *o == ':'? '\0' : *o); c++, o++)
			if (c == end)
				goto failed;
	}

	if (*o == ':')
		if (para_atoi32(++o, port) < 0 ||
		    *port < 0 || *port > 0xffff)
			goto failed;

	if (host_string_ok(host))
		return host;
failed:
	*host = '\0';
	return NULL;
}

/**
 * Stringify port number, resolve into service name where defined.
 * \param port 2-byte port number, in host-byte-order.
 * \param transport Transport protocol name (e.g. "udp", "tcp"), or NULL.
 * \return Pointer to static result buffer.
 *
 * \sa getservent(3), services(5), nsswitch.conf(5)
 */
const char *stringify_port(int port, const char *transport)
{
	static char service[NI_MAXSERV];

	if (port < 0 || port > 0xFFFF) {
		snprintf(service, sizeof(service), "undefined (%d)", port);
	} else {
		struct servent *se = getservbyport(htons(port), transport);

		if (se == NULL)
			snprintf(service, sizeof(service), "%u", port);
		else
			snprintf(service, sizeof(service), "%s", se->s_name);
	}
	return service;
}

/**
 * Determine the socket type for a given layer-4 protocol.
 *
 * \param l4type The symbolic name of the transport-layer protocol.
 *
 * \sa ip(7), socket(2)
 */
static inline int sock_type(const unsigned l4type)
{
	switch (l4type) {
	case IPPROTO_UDP:	return SOCK_DGRAM;
	case IPPROTO_TCP:	return SOCK_STREAM;
	case IPPROTO_DCCP:	return SOCK_DCCP;
	}
	return -1;		/* not supported here */
}

/**
 * Pretty-print transport-layer name.
 */
static const char *layer4_name(const unsigned l4type)
{
	switch (l4type) {
	case IPPROTO_UDP:       return "UDP";
	case IPPROTO_TCP:       return "TCP";
	case IPPROTO_DCCP:      return "DCCP";
	}
	return "UNKNOWN PROTOCOL";
}

/**
 * Flowopts: Transport-layer independent encapsulation of socket options.
 *
 * These collect individual socket options into a queue, which is disposed of
 * directly after makesock(). The 'pre_conn_opt' structure is for internal use
 * only and should not be visible elsewhere.
 *
 * \sa setsockopt(2), makesock()
 */
struct pre_conn_opt {
	int		sock_level;	/**< Second argument to setsockopt() */
	int		sock_option;	/**< Third argument to setsockopt()  */
	char		*opt_name;	/**< Stringified \a sock_option	     */
	void		*opt_val;	/**< Fourth argument to setsockopt() */
	socklen_t	opt_len;	/**< Fifth argument to setsockopt()  */

	struct list_head node;		/**< FIFO, as sockopt order matters. */
};

/** FIFO list of pre-connection socket options to be set */
struct flowopts {
	struct list_head sockopts;
};

struct flowopts *flowopt_new(void)
{
	struct flowopts *new = para_malloc(sizeof(*new));

	INIT_LIST_HEAD(&new->sockopts);
	return new;
}

/**
 * Append new socket option to flowopt queue.
 *
 * \param fo   The flowopt queue to append to.
 * \param lev  Level at which \a opt resides.
 * \param opt  New option to add.
 * \param name Stringified name of \a opt.
 * \param val  The value to set \a opt to.
 * \param len  Length of \a val.
 *
 *  \sa setsockopt(2)
 */
void flowopt_add(struct flowopts *fo, int lev, int opt,
		 char *name, const void *val, int len)
{
	struct pre_conn_opt *new = para_malloc(sizeof(*new));

	new->sock_option = opt;
	new->sock_level	 = lev;
	new->opt_name	 = para_strdup(name);

	if (val == NULL) {
		new->opt_val = NULL;
		new->opt_len = 0;
	} else {
		new->opt_val = para_malloc(len);
		new->opt_len = len;
		memcpy(new->opt_val, val, len);
	}

	list_add_tail(&new->node, &fo->sockopts);
}

void flowopt_add_bool(struct flowopts *fo, int lev, int opt,
		      char *optname, bool on_or_off)
{
	int on = on_or_off;	/* kernel takes 'int' */

	flowopt_add(fo, lev, opt, optname, &on, sizeof(on));
}

/** Set the entire bunch of pre-connection options at once. */
static void flowopt_setopts(int sockfd, struct flowopts *fo)
{
	struct pre_conn_opt *pc;

	if (fo == NULL)
		return;

	list_for_each_entry(pc, &fo->sockopts, node)
		if (setsockopt(sockfd, pc->sock_level, pc->sock_option,
				       pc->opt_val, pc->opt_len) < 0) {
			PARA_EMERG_LOG("Can not set %s socket option: %s",
				       pc->opt_name, strerror(errno));
			exit(EXIT_FAILURE);
		}
}

static void flowopt_cleanup(struct flowopts *fo)
{
	struct pre_conn_opt *cur, *next;

	if (fo == NULL)
		return;

	list_for_each_entry_safe(cur, next, &fo->sockopts, node) {
		free(cur->opt_name);
		free(cur->opt_val);
		free(cur);
	}
	free(fo);
}

/**
 * Resolve IPv4/IPv6 address and create a ready-to-use active or passive socket.
 *
 * \param l4type The layer-4 type (\p IPPROTO_xxx).
 * \param passive Whether this is a passive (1) or active (0) socket.
 * \param host Remote or local hostname or IPv/6 address string.
 * \param port_number Decimal port number.
 * \param fo Socket options to be set before making the connection.
 *
 * This creates a ready-made IPv4/v6 socket structure after looking up the
 * necessary parameters. The interpretation of \a host depends on the value of
 * \a passive:
 *	- on a passive socket host is interpreted as an interface IPv4/6 address
 *	  (can be left NULL);
 *	- on an active socket, \a host is the peer DNS name or IPv4/6 address
 *	  to connect to;
 * 	- \a port_number is in either case the numeric port number (not service
 * 	  string).
 *
 * Furthermore, bind(2) is called on passive sockets, and connect(2) on active
 * sockets. The algorithm tries all possible address combinations until it
 * succeeds. If \a fo is supplied, options are set and cleanup is performed.
 *
 * \return This function returns 1 on success and \a -E_ADDRESS_LOOKUP when no
 * matching connection could be set up (with details in the error log).
 *
 *  \sa ipv6(7), getaddrinfo(3), bind(2), connect(2).
 */
int makesock(unsigned l4type, bool passive,
	     const char *host, uint16_t port_number,
	     struct flowopts *fo)
{
	struct addrinfo *local = NULL, *src,
			*remote = NULL, *dst, hints;
	unsigned int	l3type = AF_UNSPEC;
	int		rc, on = 1, sockfd = -1,
			socktype = sock_type(l4type);
	char port[6]; /* port number has at most 5 digits */

	sprintf(port, "%u", port_number);
	/* Set up address hint structure */
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = l3type;
	hints.ai_socktype = socktype;
	/* 
	 * getaddrinfo does not support SOCK_DCCP, so for the sake of lookup
	 * (and only then) pretend to be UDP.
	 */
	if (l4type == IPPROTO_DCCP)
		hints.ai_socktype = SOCK_DGRAM;

	/* only use addresses available on the host */
	hints.ai_flags = AI_ADDRCONFIG;
	if (l3type == AF_INET6)
		/* use v4-mapped-v6 if no v6 addresses found */
		hints.ai_flags |= AI_V4MAPPED | AI_ALL;

	if (passive && host == NULL)
		hints.ai_flags |= AI_PASSIVE;

	/* Obtain local/remote address information */
	if ((rc = getaddrinfo(host, port, &hints, passive ? &local : &remote))) {
		PARA_ERROR_LOG("can not resolve %s address %s#%s: %s.\n",
				layer4_name(l4type),
				host? host : (passive? "[loopback]" : "[localhost]"),
				port, gai_strerror(rc));
		return -E_ADDRESS_LOOKUP;
	}

	/* Iterate over all src/dst combination, exhausting dst first */
	for (src = local, dst = remote; src != NULL || dst != NULL; /* no op */ ) {
		if (src && dst && src->ai_family == AF_INET
				&& dst->ai_family == AF_INET6)
			goto get_next_dst; /* v4 -> v6 is not possible */

		sockfd = socket(src ? src->ai_family : dst->ai_family,
			socktype, l4type);
		if (sockfd < 0)
			goto get_next_dst;

		/*
		 * Reuse the address on passive sockets to avoid failure on
		 * restart (protocols using listen()) and when creating
		 * multiple listener instances (UDP multicast).
		 */
		if (passive && setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR,
						  &on, sizeof(on)) == -1) {
			rc = errno;
			close(sockfd);
			PARA_ERROR_LOG("can not set SO_REUSEADDR: %s\n",
				       strerror(rc));
			return -ERRNO_TO_PARA_ERROR(rc);
		}
		flowopt_setopts(sockfd, fo);

		if (src) {
			if (bind(sockfd, src->ai_addr, src->ai_addrlen) < 0) {
				close(sockfd);
				goto get_next_src;
			}
			if (!dst) /* bind-only completed successfully */
				break;
		}

		if (dst && connect(sockfd, dst->ai_addr, dst->ai_addrlen) == 0)
			break; /* connection completed successfully */
		close(sockfd);
get_next_dst:
		if (dst && (dst = dst->ai_next))
			continue;
get_next_src:
		if (src && (src = src->ai_next)) /* restart inner loop */
			dst = remote;
	}
	if (local)
		freeaddrinfo(local);
	if (remote)
		freeaddrinfo(remote);
	flowopt_cleanup(fo);

	if (src == NULL && dst == NULL) {
		rc = errno;
		PARA_ERROR_LOG("can not create %s socket %s#%s.\n",
			layer4_name(l4type), host? host : (passive?
			"[loopback]" : "[localhost]"), port);
		return -ERRNO_TO_PARA_ERROR(rc);
	}
	return sockfd;
}

/**
 * Create a passive / listening socket.
 *
 * \param l4type The transport-layer type (\p IPPROTO_xxx).
 * \param port The decimal port number to listen on.
 * \param fo Flowopts (if any) to set before starting to listen.
 *
 * \return Positive integer (socket descriptor) on success, negative value
 * otherwise.
 *
 * \sa makesock(), ip(7), ipv6(7), bind(2), listen(2).
 */
int para_listen(unsigned l4type, uint16_t port, struct flowopts *fo)
{
	int ret, fd = makesock(l4type, 1, NULL, port, fo);

	if (fd > 0) {
		ret = listen(fd, BACKLOG);
		if (ret < 0) {
			ret = errno;
			close(fd);
			return -ERRNO_TO_PARA_ERROR(ret);
		}
		PARA_INFO_LOG("listening on %s port %u, fd %d\n",
			layer4_name(l4type), port, fd);
	}
	return fd;
}

/**
 * Determine IPv4/v6 socket address length.
 * \param sa Container of IPv4 or IPv6 address.
 * \return Address-family dependent address length.
 */
static socklen_t salen(const struct sockaddr *sa)
{
	assert(sa->sa_family == AF_INET || sa->sa_family == AF_INET6);

	return sa->sa_family == AF_INET6
		? sizeof(struct sockaddr_in6)
		: sizeof(struct sockaddr_in);
}

/**
 * Process IPv4/v6 address, turn v6-mapped-v4 address into normal IPv4 address.
 * \param ss Container of IPv4/6 address.
 * \return Pointer to normalized address (may be static storage).
 *
 * \sa RFC 3493
 */
static const struct sockaddr *
normalize_ip_address(const struct sockaddr_storage *ss)
{
	const struct sockaddr_in6 *ia6 = (const struct sockaddr_in6 *)ss;

	assert(ss->ss_family == AF_INET || ss->ss_family == AF_INET6);

	if (ss->ss_family == AF_INET6 && IN6_IS_ADDR_V4MAPPED(&ia6->sin6_addr)) {
		static struct sockaddr_in ia;

		ia.sin_family = AF_INET;
		ia.sin_port   = ia6->sin6_port;
		memcpy(&ia.sin_addr.s_addr, &(ia6->sin6_addr.s6_addr[12]), 4);
		return (const struct sockaddr *)&ia;
	}
	return (const struct sockaddr *)ss;
}

/**
 * Print numeric host and port number (beware - uses static char).
 *
 * \param sa The IPv4/IPv6 socket address to use.
 *
 * \sa getnameinfo(3), services(5), nsswitch.conf(5)
 */
static char *host_and_port(const struct sockaddr_storage *ss)
{
	const struct sockaddr *sa = normalize_ip_address(ss);
	char hbuf[NI_MAXHOST], sbuf[NI_MAXSERV];
	static char output[sizeof(hbuf) + sizeof(sbuf) + 2];
	int ret;

	ret = getnameinfo(sa, salen(sa),
			  hbuf, sizeof(hbuf),
			  sbuf, sizeof(sbuf),
			  NI_NUMERICHOST);
	if (ret == 0) {
		snprintf(output, sizeof(output), "%s#%s", hbuf, sbuf);
	} else {
		snprintf(output, sizeof(output), "(unknown)");
		PARA_WARNING_LOG("hostname lookup error (%s).\n",
				 gai_strerror(ret));
	}
	return output;
}

/**
 * Look up the local or remote side of a connected socket structure.
 *
 * \param fd The socket descriptor of the connected socket.
 * \param getname Either \p getsockname() for local, or \p getpeername() for
 * remote side.
 *
 * \return A static character string identifying hostname and port of the
 * chosen side.
 *
 * \sa getsockname(2), getpeername(2).
 */
static char *__get_sock_name(int fd, int (*getname)(int, struct sockaddr*,
		socklen_t *))
{
	struct sockaddr_storage ss;
	socklen_t sslen = sizeof(ss);

	if (getname(fd, (struct sockaddr *)&ss, &sslen) < 0) {
		static char *dont_know = "(don't know)";
		PARA_ERROR_LOG("can not determine address from fd %d: %s\n",
			fd, strerror(errno));
		return dont_know;
	}
	return host_and_port(&ss);
}

/**
 * Look up the local side of a connected socket structure.
 *
 * \param sockfd The file descriptor of the socket.
 *
 * \return A pointer to a static buffer containing hostname an port. This
 * buffer must not be freed by the caller.
 *
 * \sa remote_name().
 */
char *local_name(int sockfd)
{
	return __get_sock_name(sockfd, getsockname);
}

/**
 * Look up the remote side of a connected socket structure.
 *
 * \param sockfd The file descriptor of the socket.
 *
 * \return Analogous to the return value of \ref local_name() but for the
 * remote side.
 *
 * \sa local_name().
 */
char *remote_name(int sockfd)
{
	return __get_sock_name(sockfd, getpeername);
}

/**
 * Extract IPv4 or IPv6-mapped-IPv4 address from sockaddr_storage.
 * \param ss Container of IPv4/6 address
 * \return Extracted IPv4 address (different from 0) or 0 if unsuccessful.
 *
 * \sa RFC 3493
 */
struct in_addr extract_v4_addr(const struct sockaddr_storage *ss)
{
	struct in_addr ia = {.s_addr = 0};
	const struct sockaddr *sa = normalize_ip_address(ss);

	if (sa->sa_family == AF_INET)
		ia = ((struct sockaddr_in *)sa)->sin_addr;
	return ia;
}

/**
 * Send a binary buffer.
 *
 * \param fd The file descriptor.
 * \param buf The buffer to be sent.
 * \param len The length of \a buf.
 *
 * Send out the buffer and try to resend the remaining part in case of short
 * writes.
 *
 * \return Standard.
 */
int send_bin_buffer(int fd, const char *buf, size_t len)
{
	if (!len)
		PARA_CRIT_LOG("len == 0\n");
	return write_all(fd, buf, &len);
}

/**
 * Send a \p NULL-terminated buffer.
 *
 * \param fd The file descriptor.
 * \param buf The null-terminated buffer to be send.
 *
 * This is equivalent to send_bin_buffer(fd, buf, strlen(buf)).
 *
 * \return Standard.
 */
int send_buffer(int fd, const char *buf)
{
	return send_bin_buffer(fd, buf, strlen(buf));
}

/**
 * Send a buffer given by a format string.
 *
 * \param fd The file descriptor.
 * \param fmt A format string.
 *
 * \return Standard.
 */
__printf_2_3 int send_va_buffer(int fd, const char *fmt, ...)
{
	char *msg;
	int ret;

	PARA_VSPRINTF(fmt, msg);
	ret = send_buffer(fd, msg);
	free(msg);
	return ret;
}

/**
 * Receive data from a file descriptor.
 *
 * \param fd The file descriptor.
 * \param buf The buffer to write the data to.
 * \param size The size of \a buf.
 *
 * Receive at most \a size bytes from file descriptor \a fd.
 *
 * \return The number of bytes received on success, negative on errors, zero if
 * the peer has performed an orderly shutdown.
 *
 * \sa recv(2).
 */
__must_check int recv_bin_buffer(int fd, char *buf, size_t size)
{
	ssize_t n;

	n = recv(fd, buf, size, 0);
	if (n == -1)
		return -ERRNO_TO_PARA_ERROR(errno);
	return n;
}

/**
 * Receive and write terminating NULL byte.
 *
 * \param fd The file descriptor.
 * \param buf The buffer to write the data to.
 * \param size The size of \a buf.
 *
 * Read at most \a size - 1 bytes from file descriptor \a fd and
 * write a NULL byte at the end of the received data.
 *
 * \return The return value of the underlying call to \a recv_bin_buffer().
 *
 * \sa recv_bin_buffer()
 */
int recv_buffer(int fd, char *buf, size_t size)
{
	int n;

	assert(size);
	n = recv_bin_buffer(fd, buf, size - 1);
	if (n >= 0)
		buf[n] = '\0';
	else
		*buf = '\0';
	return n;
}

/**
 * Wrapper around the accept system call.
 *
 * \param fd The listening socket.
 * \param addr Structure which is filled in with the address of the peer socket.
 * \param size Should contain the size of the structure pointed to by \a addr.
 *
 * Accept incoming connections on \a addr. Retry if interrupted.
 *
 * \return The new file descriptor on success, negative on errors.
 *
 * \sa accept(2).
 */
int para_accept(int fd, void *addr, socklen_t size)
{
	int new_fd;

	do
		new_fd = accept(fd, (struct sockaddr *) addr, &size);
	while (new_fd < 0 && errno == EINTR);
	return new_fd < 0? -ERRNO_TO_PARA_ERROR(errno) : new_fd;
}

/**
 * Probe the list of DCCP CCIDs configured on this host.
 * \param ccid_array Pointer to return statically allocated array in.
 * \return Number of elements returned in \a ccid_array or error.
 *
 * NB: This feature is only available on Linux > 2.6.30; on older kernels
 * ENOPROTOOPT ("Protocol not available") will be returned.
 */
int dccp_available_ccids(uint8_t **ccid_array)
{
	static uint8_t ccids[DCCP_MAX_HOST_CCIDS];
	socklen_t nccids = sizeof(ccids);
	int ret, fd;

	ret = fd = makesock(IPPROTO_DCCP, 1, NULL, 0, NULL);
	if (ret < 0)
		return ret;

	if (getsockopt(fd, SOL_DCCP, DCCP_SOCKOPT_AVAILABLE_CCIDS,
						ccids, &nccids) < 0) {
		ret = errno;
		close(fd);
		PARA_ERROR_LOG("No DCCP_SOCKOPT_AVAILABLE_CCIDS: %s\n",
				strerror(ret));
		return -ERRNO_TO_PARA_ERROR(ret);
	}

	close(fd);
	*ccid_array = ccids;
	return nccids;
}

/**
 * Prepare a structure for \p AF_UNIX socket addresses.
 *
 * \param u Pointer to the struct to be prepared.
 * \param name The socket pathname.
 *
 * This just copies \a name to the sun_path component of \a u.
 *
 * \return Positive on success, \p -E_NAME_TOO_LONG if \a name is longer
 * than \p UNIX_PATH_MAX.
 */
static int init_unix_addr(struct sockaddr_un *u, const char *name)
{
	if (strlen(name) >= UNIX_PATH_MAX)
		return -E_NAME_TOO_LONG;
	memset(u->sun_path, 0, UNIX_PATH_MAX);
	u->sun_family = PF_UNIX;
	strcpy(u->sun_path, name);
	return 1;
}

/**
 * Prepare, create, and bind a socket for local communication.
 *
 * \param name The socket pathname.
 * \param unix_addr Pointer to the \p AF_UNIX socket structure.
 * \param mode The desired mode of the socket.
 *
 * This function creates a local socket for sequenced, reliable,
 * two-way, connection-based byte streams.
 *
 * \return The file descriptor, on success, negative on errors.
 *
 * \sa socket(2)
 * \sa bind(2)
 * \sa chmod(2)
 */
int create_local_socket(const char *name, struct sockaddr_un *unix_addr,
		mode_t mode)
{
	int fd, ret;

	ret = init_unix_addr(unix_addr, name);
	if (ret < 0)
		return ret;
	ret = socket(PF_UNIX, SOCK_STREAM, 0);
	if (ret < 0)
		return -ERRNO_TO_PARA_ERROR(errno);
	fd = ret;
	ret = bind(fd, (struct sockaddr *) unix_addr, UNIX_PATH_MAX);
	if (ret < 0) {
		ret = -ERRNO_TO_PARA_ERROR(errno);
		goto err;
	}
	ret = -E_CHMOD;
	if (chmod(name, mode) < 0)
		goto err;
	return fd;
err:
	close(fd);
	return ret;
}

/**
 * Prepare, create, and connect to a Unix domain socket for local communication.
 *
 * \param name The socket pathname.
 *
 * This function creates a local socket for sequenced, reliable, two-way,
 * connection-based byte streams.
 *
 * \return The file descriptor, on success, negative on errors.
 *
 * \sa create_local_socket(), unix(7), connect(2).
 */
int create_remote_socket(const char *name)
{
	struct sockaddr_un unix_addr;
	int fd, ret;

	ret = init_unix_addr(&unix_addr, name);
	if (ret < 0)
		return ret;
	fd = socket(PF_UNIX, SOCK_STREAM, 0);
	if (fd < 0)
		return -ERRNO_TO_PARA_ERROR(errno);
	if (connect(fd, (struct sockaddr *)&unix_addr, sizeof(unix_addr)) == -1) {
		ret = -ERRNO_TO_PARA_ERROR(errno);
		goto err;
	}
	return fd;
err:
	close(fd);
	return ret;
}

#ifndef HAVE_UCRED
ssize_t send_cred_buffer(int sock, char *buf)
{
	return send_buffer(sock, buf);
}
int recv_cred_buffer(int fd, char *buf, size_t size)
{
	return recv_buffer(fd, buf, size) > 0? 1 : -E_RECVMSG;
}
#else /* HAVE_UCRED */
/**
 * Send \p NULL-terminated buffer and Unix credentials of the current process.
 *
 * \param sock The socket file descriptor.
 * \param buf The buffer to be sent.
 *
 * \return On success, this call returns the number of characters sent.  On
 * error, \p -E_SENDMSG is returned.
 *
 * \sa sendmsg(2), okir's Black Hats Manual.
 */
ssize_t send_cred_buffer(int sock, char *buf)
{
	char control[sizeof(struct cmsghdr) + sizeof(struct ucred)];
	struct msghdr msg;
	struct cmsghdr *cmsg;
	static struct iovec iov;
	struct ucred c;
	int ret;

	/* Response data */
	iov.iov_base = buf;
	iov.iov_len  = strlen(buf);
	c.pid = getpid();
	c.uid = getuid();
	c.gid = getgid();
	/* compose the message */
	memset(&msg, 0, sizeof(msg));
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_control = control;
	msg.msg_controllen = sizeof(control);
	/* attach the ucred struct */
	cmsg = CMSG_FIRSTHDR(&msg);
	cmsg->cmsg_level = SOL_SOCKET;
	cmsg->cmsg_type = SCM_CREDENTIALS;
	cmsg->cmsg_len = CMSG_LEN(sizeof(struct ucred));
	*(struct ucred *)CMSG_DATA(cmsg) = c;
	msg.msg_controllen = cmsg->cmsg_len;
	ret = sendmsg(sock, &msg, 0);
	if (ret  < 0)
		ret = -E_SENDMSG;
	return ret;
}

static void dispose_fds(int *fds, unsigned num)
{
	int i;

	for (i = 0; i < num; i++)
		close(fds[i]);
}

/**
 * Receive a buffer and the Unix credentials of the sending process.
 *
 * \param fd the socket file descriptor.
 * \param buf the buffer to store the message.
 * \param size the size of \a buffer.
 *
 * \return negative on errors, the user id on success.
 *
 * \sa recvmsg(2), okir's Black Hats Manual.
 */
int recv_cred_buffer(int fd, char *buf, size_t size)
{
	char control[255];
	struct msghdr msg;
	struct cmsghdr *cmsg;
	struct iovec iov;
	int result = 0;
	int yes = 1;
	struct ucred cred;

	setsockopt(fd, SOL_SOCKET, SO_PASSCRED, &yes, sizeof(int));
	memset(&msg, 0, sizeof(msg));
	memset(buf, 0, size);
	iov.iov_base = buf;
	iov.iov_len = size;
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_control = control;
	msg.msg_controllen = sizeof(control);
	if (recvmsg(fd, &msg, 0) < 0)
		return -E_RECVMSG;
	result = -E_SCM_CREDENTIALS;
	cmsg = CMSG_FIRSTHDR(&msg);
	while (cmsg) {
		if (cmsg->cmsg_level == SOL_SOCKET && cmsg->cmsg_type
				== SCM_CREDENTIALS) {
			memcpy(&cred, CMSG_DATA(cmsg), sizeof(struct ucred));
			result = cred.uid;
		} else
			if (cmsg->cmsg_level == SOL_SOCKET
					&& cmsg->cmsg_type == SCM_RIGHTS) {
				dispose_fds((int *) CMSG_DATA(cmsg),
					(cmsg->cmsg_len - CMSG_LEN(0))
					/ sizeof(int));
			}
		cmsg = CMSG_NXTHDR(&msg, cmsg);
	}
	return result;
}
#endif /* HAVE_UCRED */

/**
 * Receive a buffer and check for a pattern.
 *
 * \param fd The file descriptor to receive from.
 * \param pattern The expected pattern.
 * \param bufsize The size of the internal buffer.
 *
 * \return Positive if \a pattern was received, negative otherwise.
 *
 * This function tries to receive at most \a bufsize bytes from file descriptor
 * \a fd. If at least \p strlen(\a pattern) bytes were received, the beginning
 * of the received buffer is compared with \a pattern, ignoring case.
 *
 * \sa recv_buffer(), \sa strncasecmp(3).
 */
int recv_pattern(int fd, const char *pattern, size_t bufsize)
{
	size_t len = strlen(pattern);
	char *buf = para_malloc(bufsize + 1);
	int ret = -E_RECV_PATTERN, n = recv_buffer(fd, buf, bufsize + 1);

	if (n < len)
		goto out;
	if (strncasecmp(buf, pattern, len))
		goto out;
	ret = 1;
out:
	if (ret < 0) {
		PARA_NOTICE_LOG("did not receive pattern '%s'\n", pattern);
		if (n > 0)
			PARA_NOTICE_LOG("recvd %d bytes: %s\n", n, buf);
		else if (n < 0)
			PARA_NOTICE_LOG("%s\n", para_strerror(-n));
	}
	free(buf);
	return ret;
}
